@@ -17,24 +17,31 @@ import uuid
 import warnings
 
 from datetime import datetime
-from pprint import pformat
+from inspect import isclass
+from random import Random
 from types import FunctionType
+from threading import local
 
 if sys.version_info >= (3, 2):
     import contextlib
 else:
     import contextlib2 as contextlib
 
+try:
+    from thread import get_ident as get_thread_ident
+except ImportError:
+    from _thread import get_ident as get_thread_ident
+
 import raven
 from raven.conf import defaults
 from raven.conf.remote import RemoteConfig
-from raven.context import Context
 from raven.exceptions import APIError, RateLimited
 from raven.utils import json, get_versions, get_auth_header, merge_dicts
-from raven._compat import text_type, iteritems
+from raven.utils.compat import text_type, iteritems
 from raven.utils.encoding import to_unicode
 from raven.utils.serializer import transform
-from raven.utils.stacks import get_stack_info, iter_stack_frames, get_culprit
+from raven.utils.stacks import get_stack_info, iter_stack_frames
+from raven.utils.transaction import TransactionStack
 from raven.transport.registry import TransportRegistry, default_transports
 
 # enforce imports to avoid obscure stacktraces with MemoryError
@@ -47,9 +54,21 @@ __excepthook__ = None
 
 PLATFORM_NAME = 'python'
 
+SDK_VALUE = {
+    'name': 'raven-python',
+    'version': raven.VERSION,
+}
+
 # singleton for the client
 Raven = None
 
+if sys.version_info >= (3, 2):
+    basestring = str
+
+
+def get_excepthook_client():
+    return getattr(sys.excepthook, 'raven_client', None)
+
 
 class ModuleProxyCache(dict):
     def __missing__(self, key):
@@ -122,17 +141,22 @@ class Client(object):
     >>>     ident = client.get_ident(client.captureException())
     >>>     print "Exception caught; reference is %s" % ident
     """
+
     logger = logging.getLogger('raven')
     protocol_version = '6'
 
     _registry = TransportRegistry(transports=default_transports)
 
     def __init__(self, dsn=None, raise_send_errors=False, transport=None,
-                 install_sys_hook=True, **options):
+                 install_sys_hook=True, install_logging_hook=True,
+                 hook_libraries=None, enable_breadcrumbs=True,
+                 _random_seed=None, **options):
         global Raven
 
         o = options
 
+        self._local_state = local()
+
         self.raise_send_errors = raise_send_errors
 
         # configure loggers first
@@ -148,7 +172,9 @@ class Client(object):
 
         self.include_paths = set(o.get('include_paths') or [])
         self.exclude_paths = set(o.get('exclude_paths') or [])
-        self.name = text_type(o.get('name') or o.get('machine') or defaults.NAME)
+        self.name = text_type(
+            o.get('name') or os.environ.get('SENTRY_NAME') or
+            o.get('machine') or defaults.NAME)
         self.auto_log_stacks = bool(
             o.get('auto_log_stacks') or defaults.AUTO_LOG_STACKS)
         self.capture_locals = bool(
@@ -160,19 +186,33 @@ class Client(object):
         self.site = o.get('site')
         self.include_versions = o.get('include_versions', True)
         self.processors = o.get('processors')
+        self.sanitize_keys = o.get('sanitize_keys')
         if self.processors is None:
             self.processors = defaults.PROCESSORS
 
         context = o.get('context')
         if context is None:
-            context = {'sys.argv': sys.argv[:]}
+            context = {'sys.argv': getattr(sys, 'argv', [])[:]}
         self.extra = context
         self.tags = o.get('tags') or {}
-        self.environment = o.get('environment') or None
-        self.release = o.get('release') or os.environ.get('HEROKU_SLUG_COMMIT')
+        self.environment = (
+            o.get('environment') or os.environ.get('SENTRY_ENVIRONMENT', None))
+        self.release = (
+            o.get('release') or os.environ.get('SENTRY_RELEASE') or
+            os.environ.get('HEROKU_SLUG_COMMIT'))
+        self.repos = self._format_repos(o.get('repos'))
+        self.sample_rate = (
+            o.get('sample_rate')
+            if o.get('sample_rate') is not None
+            else 1
+        )
+        self.transaction = TransactionStack()
+        self.ignore_exceptions = set(o.get('ignore_exceptions') or ())
 
         self.module_cache = ModuleProxyCache()
 
+        self._random = Random(_random_seed)
+
         if not self.is_enabled():
             self.logger.info(
                 'Raven is not configured (logging is disabled). Please see the'
@@ -181,19 +221,40 @@ class Client(object):
         if Raven is None:
             Raven = self
 
-        self._context = Context()
+        # We want to remember the creating thread id here because this
+        # comes in useful for the context special handling
+        self.main_thread_id = get_thread_ident()
+        self.enable_breadcrumbs = enable_breadcrumbs
+
+        from raven.context import Context
+        self._context = Context(self)
 
         if install_sys_hook:
             self.install_sys_hook()
 
+        if install_logging_hook:
+            self.install_logging_hook()
+
+        self.hook_libraries(hook_libraries)
+
+    def _format_repos(self, value):
+        result = {}
+        if value:
+            for path, config in iteritems(value):
+                if path[0] != '/':
+                    # assume its a module
+                    path = os.path.abspath(__import__(path).__file__)
+                result[path] = config
+        return result
+
     def set_dsn(self, dsn=None, transport=None):
-        if dsn is None and os.environ.get('SENTRY_DSN'):
+        if not dsn and os.environ.get('SENTRY_DSN'):
             msg = "Configuring Raven from environment variable 'SENTRY_DSN'"
             self.logger.debug(msg)
             dsn = os.environ['SENTRY_DSN']
 
         if dsn not in self._transport_cache:
-            if dsn is None:
+            if not dsn:
                 result = RemoteConfig(transport=transport)
             else:
                 result = RemoteConfig.from_string(
@@ -215,10 +276,19 @@ class Client(object):
             __excepthook__ = sys.excepthook
 
         def handle_exception(*exc_info):
-            self.captureException(exc_info=exc_info)
+            self.captureException(exc_info=exc_info, level='fatal')
             __excepthook__(*exc_info)
+        handle_exception.raven_client = self
         sys.excepthook = handle_exception
 
+    def install_logging_hook(self):
+        from raven.breadcrumbs import install_logging_hook
+        install_logging_hook()
+
+    def hook_libraries(self, libraries):
+        from raven.breadcrumbs import hook_libraries
+        hook_libraries(libraries)
+
     @classmethod
     def register_scheme(cls, scheme, transport_class):
         cls._registry.register_scheme(scheme, transport_class)
@@ -265,18 +335,17 @@ class Client(object):
         >>> # Specify a scheme to use (http or https)
         >>> print client.get_public_dsn('https')
         """
-        if not self.is_enabled():
-            return
-        url = self.remote.get_public_dsn()
-        if not scheme:
+        if self.is_enabled():
+            url = self.remote.get_public_dsn()
+            if scheme:
+                return '%s:%s' % (scheme, url)
+
             return url
-        return '%s:%s' % (scheme, url)
 
     def _get_exception_key(self, exc_info):
         # On certain celery versions the tb_frame attribute might
         # not exist or be `None`.
-        code_id = 0
-        last_id = 0
+        code_id = last_id = 0
         try:
             code_id = id(exc_info[2] and exc_info[2].tb_frame.f_code)
             last_id = exc_info[2] and exc_info[2].tb_lasti or 0
@@ -307,7 +376,6 @@ class Client(object):
         The result of ``build_msg`` should be a standardized dict, with
         all default values available.
         """
-
         # create ID client-side so that it can be passed to application
         event_id = uuid.uuid4().hex
 
@@ -371,13 +439,9 @@ class Client(object):
                         not any(path.startswith(x) for x in self.exclude_paths)
                     )
 
+        transaction = None
         if not culprit:
-            if 'stacktrace' in data:
-                culprit = get_culprit(data['stacktrace']['frames'])
-            elif 'exception' in data:
-                stacktrace = data['exception']['values'][0].get('stacktrace')
-                if stacktrace:
-                    culprit = get_culprit(stacktrace['frames'])
+            transaction = self.transaction.peek()
 
         if not data.get('level'):
             data['level'] = kwargs.get('level') or logging.ERROR
@@ -402,7 +466,9 @@ class Client(object):
         if site:
             data['tags'].setdefault('site', site)
 
-        if culprit:
+        if transaction:
+            data['transaction'] = transaction
+        elif culprit:
             data['culprit'] = culprit
 
         if fingerprint:
@@ -429,6 +495,20 @@ class Client(object):
         data.setdefault('time_spent', time_spent)
         data.setdefault('event_id', event_id)
         data.setdefault('platform', PLATFORM_NAME)
+        data.setdefault('sdk', SDK_VALUE)
+        data.setdefault('repos', self.repos)
+
+        # insert breadcrumbs
+        if self.enable_breadcrumbs:
+            crumbs = self.context.breadcrumbs.get_buffer()
+            if crumbs:
+                # Make sure we send the crumbs here as "values" as we use the
+                # raven client internally in sentry and the alternative
+                # submission option of a list here is not supported by the
+                # internal sender.
+                data.setdefault('breadcrumbs', {
+                    'values': crumbs
+                })
 
         return data
 
@@ -486,13 +566,15 @@ class Client(object):
         Update the tags context for future events.
 
         >>> client.tags_context({'version': '1.0'})
+
         """
         return self.context.merge({
             'tags': data,
         })
 
     def capture(self, event_type, data=None, date=None, time_spent=None,
-                extra=None, stack=None, tags=None, **kwargs):
+                extra=None, stack=None, tags=None, sample_rate=None,
+                **kwargs):
         """
         Captures and processes an event and pipes it off to SentryClient.send.
 
@@ -540,9 +622,9 @@ class Client(object):
         :param extra: a dictionary of additional standard metadata
         :param stack: a stacktrace for the event
         :param tags: dict of extra tags
-        :return: a tuple with a 32-length string identifying this event
+        :param sample_rate: a float in the range [0, 1] to sample this message
+        :return: a 32-length string identifying this event
         """
-
         if not self.is_enabled():
             return
 
@@ -550,13 +632,25 @@ class Client(object):
         if exc_info is not None:
             if self.skip_error_for_logging(exc_info):
                 return
+            elif not self.should_capture(exc_info):
+                self.logger.info(
+                    'Not capturing exception due to filters: %s', exc_info[0],
+                    exc_info=sys.exc_info())
+                return
             self.record_exception_seen(exc_info)
 
         data = self.build_msg(
             event_type, data, date, time_spent, extra, stack, tags=tags,
             **kwargs)
 
-        self.send(**data)
+        # should this event be sampled?
+        if sample_rate is None:
+            sample_rate = self.sample_rate
+
+        if self._random.random() < sample_rate:
+            self.send(**data)
+
+        self._local_state.last_event_id = data['event_id']
 
         return data['event_id']
 
@@ -572,7 +666,7 @@ class Client(object):
             for frame in data['stacktrace']['frames']:
                 yield frame
         if 'exception' in data:
-            for frame in data['exception']['values'][0]['stacktrace']['frames']:
+            for frame in data['exception']['values'][-1]['stacktrace'].get('frames', []):
                 yield frame
 
     def _successful_send(self):
@@ -588,9 +682,10 @@ class Client(object):
                 type(exc).__name__, exc.message)
         else:
             self.error_logger.error(
-                'Sentry responded with an error: %s (url: %s)\n%s',
-                exc, url, pformat(data),
-                exc_info=True
+                'Sentry responded with an error: %s (url: %s)',
+                exc, url,
+                exc_info=True,
+                extra={'data': data}
             )
 
         self._log_failed_submission(data)
@@ -603,13 +698,13 @@ class Client(object):
         """
         message = data.pop('message', '<no message value>')
         output = [message]
-        if 'exception' in data and 'stacktrace' in data['exception']['values'][0]:
+        if 'exception' in data and 'stacktrace' in data['exception']['values'][-1]:
             # try to reconstruct a reasonable version of the exception
-            for frame in data['exception']['values'][0]['stacktrace']['frames']:
+            for frame in data['exception']['values'][-1]['stacktrace'].get('frames', []):
                 output.append('  File "%(fn)s", line %(lineno)s, in %(func)s' % {
-                    'fn': frame['filename'],
-                    'lineno': frame['lineno'],
-                    'func': frame['function'],
+                    'fn': frame.get('filename', 'unknown_filename'),
+                    'lineno': frame.get('lineno', -1),
+                    'func': frame.get('function', 'unknown_function'),
                 })
 
         self.uncaught_logger.error(output)
@@ -633,11 +728,11 @@ class Client(object):
 
         try:
             transport = self.remote.get_transport()
-            if transport.async:
-                transport.async_send(data, headers, self._successful_send,
+            if transport.is_async:
+                transport.async_send(url, data, headers, self._successful_send,
                                      failed_send)
             else:
-                transport.send(data, headers)
+                transport.send(url, data, headers)
                 self._successful_send()
         except Exception as e:
             if self.raise_send_errors:
@@ -676,7 +771,7 @@ class Client(object):
             'Content-Type': 'application/octet-stream',
         }
 
-        self.send_remote(
+        return self.send_remote(
             url=self.remote.store_endpoint,
             data=message,
             headers=headers,
@@ -724,9 +819,26 @@ class Client(object):
         """
         if exc_info is None or exc_info is True:
             exc_info = sys.exc_info()
+
         return self.capture(
             'raven.events.Exception', exc_info=exc_info, **kwargs)
 
+    def should_capture(self, exc_info):
+        exc_type = exc_info[0]
+        exc_name = '%s.%s' % (exc_type.__module__, exc_type.__name__)
+        exclusions = self.ignore_exceptions
+        string_exclusions = (e for e in exclusions if isinstance(e, basestring))
+        wildcard_exclusions = (e for e in string_exclusions if e.endswith('*'))
+        class_exclusions = (e for e in exclusions if isclass(e))
+
+        if (exc_type in exclusions
+                or exc_type.__name__ in exclusions
+                or exc_name in exclusions
+                or any(issubclass(exc_type, e) for e in class_exclusions)
+                or any(exc_name.startswith(e[:-1]) for e in wildcard_exclusions)):
+            return False
+        return True
+
     def capture_exceptions(self, function_or_exceptions=None, **kwargs):
         """
         Wrap a function or code block in try/except and automatically call
@@ -791,8 +903,29 @@ class Client(object):
             DeprecationWarning)
         return self.context(**kwargs)
 
+    def captureBreadcrumb(self, *args, **kwargs):
+        """
+        Records a breadcrumb with the current context.  They will be
+        sent with the next event.
+        """
+        # Note: framework integration should not call this method but
+        # instead use the raven.breadcrumbs.record_breadcrumb function
+        # which will record to the correct client automatically.
+        self.context.breadcrumbs.record(*args, **kwargs)
+
+    capture_breadcrumb = captureBreadcrumb
+
+    @property
+    def last_event_id(self):
+        return getattr(self._local_state, 'last_event_id', None)
+
+    @last_event_id.setter
+    def last_event_id(self, value):
+        self._local_state.last_event_id = value
+
 
 class DummyClient(Client):
-    "Sends messages into an empty void"
+    """Sends messages into an empty void."""
+
     def send(self, **kwargs):
         return None
